import arrow
from collections import namedtuple
import datetime
from unittest import mock

from keg import current_app
from pyquery import PyQuery
import sqlalchemy as sa
from sqlalchemy_utils import ArrowType
from werkzeug.datastructures import MultiDict

from .db.utils import validate_unique_exc
from .db.mixins import DefaultColsMixin

ColumnCheck = namedtuple('ColumnCheck',
                         'name, required, fk, unique, timestamp, skip_callable_date_check')
"""Validator tuple used in ``EntityBase`` to check column spec for common cases.

:param name: Field name to check.
:param required: Non-nullable status to verify.
:param fk: Foreign key to verify, e.g. "foo.id". Multiple keys can be given as list, set, or CSV.
:param unique: If true, verifies ``unique`` kwarg passed to column.
:param timestamp: Can be ``True`` for creates and ``update`` for updates. Expects column defaults.
:param skip_callable_date_check: If true, date cols won't validate that default vals are callable.
"""
ColumnCheck.__new__.__defaults__ = (True, None, None, None, False)


class DontCare(object):  # pragma: no cover
    """A placeholder object that can conform to almost anything and is always
    equal to anything it is compared against.

    Examples:

        assert [1, 2, 3] == [1, DontCare(), 3]
    """
    __eq__ = lambda *_: True
    __ne__ = lambda *_: False
    __repr__ = lambda *_: '_'

    __getattr__ = lambda *_: DontCare()
    __getitem__ = lambda *_: DontCare()
    __call__ = lambda *a, **kw: DontCare()


class EntityBase(object):
    """Base class for data model tests.

    Class attributes:

        entity_cls: SQLAlchemy entity class to test.

        column_checks: Iterable of ColumnCheck instances to use in tests. An instance
        for every field except id is expected for verification, and a test will fail if that
        expectation is not met. Note, created/updated timestamp columns have checks
        automatically generated and do not need to be specified by the developer.

        timestamp_cols: Defaults to ('created_utc', 'updated_utc'), the timestamp columns.

        delete_all_on: When to delete all records. Default "setup_class". Can also be "setup"
        or None.
    """
    entity_cls = None
    column_checks = None
    timestamp_cols = ('created_utc', 'updated_utc')
    # Set to: 'setup_class', 'setup', or None
    delete_all_on = 'setup_class'

    @classmethod
    def setup_class(cls):
        assert current_app.db_enabled, 'app.db_enabled is False'
        assert cls.entity_cls, 'The entity_cls attribute must be set.'
        assert cls.column_checks, 'column_checks should be defined for this entity'
        if cls.timestamp_cols:
            if cls.column_checks is None:
                cls.column_checks = []
            createdts_col_name, updatedts_col_name = cls.timestamp_cols
            cls.column_checks.append(ColumnCheck(createdts_col_name, timestamp='create'))
            cls.column_checks.append(ColumnCheck(updatedts_col_name, timestamp='update'))
        if cls.delete_all_on:
            cls.ent_delete_all()

    @classmethod
    def column_check_generator(cls):
        for col_check in cls.column_checks or []:
            yield col_check
        if cls.timestamp_cols:
            createdts_col_name, updatedts_col_name = cls.timestamp_cols
            yield ColumnCheck(createdts_col_name, timestamp='create')
            yield ColumnCheck(updatedts_col_name, timestamp='update')

    @classmethod
    def ent_delete_all(cls):
        if hasattr(cls.entity_cls, 'delete_cascaded'):
            cls.entity_cls.delete_cascaded()
        else:
            cls.entity_cls.delete_all()

    def setup_method(self):
        if self.delete_all_on == 'setup':
            self.ent_delete_all()

    def test_add(self):
        self.ent_delete_all()
        o = self.entity_cls.fake()
        assert self.entity_cls.query.count() == 1
        if hasattr(self.entity_cls, 'id'):
            assert o.id

    def test_column_null(self):
        for col_check in self.column_check_generator():
            col = getattr(self.entity_cls, col_check.name)
            assert col.nullable != col_check.required, \
                'Expected column "{}" to match null requirement'.format(col.name)

    def test_column_unique(self):
        for col_check in self.column_check_generator():
            col = getattr(self.entity_cls, col_check.name)
            assert col.unique == col_check.unique, 'Expected column "{}" to have unique={}'.format(
                col.name, col_check.unique)

    def test_column_fk(self):
        for col_check in self.column_check_generator():
            col = getattr(self.entity_cls, col_check.name)
            fk_count = len(col.foreign_keys)
            if col_check.fk:
                # normalize `fk` into a set
                if isinstance(col_check.fk, str):
                    # 'foo.bar' => {'foo.bar'}
                    # 'foo.bar,baz.qux' => {'foo.bar', 'baz.qux'}
                    # 'foo.bar, baz.qux' => {'foo.bar', 'baz.qux'}
                    fk_set = {partial.strip() for partial in col_check.fk.split(',')}

                elif isinstance(col_check.fk, set):
                    fk_set = col_check.fk

                else:
                    # ['foo.bar', 'baz.qux'] => {'foo.bar', 'baz.qux'}
                    fk_set = set(col_check.fk)

                assert fk_count == len(fk_set), 'FK count does not match for {}'.format(col.name)
                assert fk_set == {fk._get_colspec() for fk in col.foreign_keys}, \
                    'FK set does not match for {}'.format(col.name)

            else:
                assert not fk_count, \
                    'Didn\'t expect column "{}" to have a foreign key'.format(col.name)

    def test_column_timestamp(self):
        for col_check in self.column_check_generator():
            if not col_check.timestamp:
                continue
            col = getattr(self.entity_cls, col_check.name)
            if not isinstance(col.type, (ArrowType, sa.DateTime)):
                raise AssertionError(
                    'Column "{}" is not a recognized timestamp type.'.format(col.name))
            assert col.default
            if col_check.timestamp == 'update':
                assert col.onupdate, 'Column "{}" should have onupdate set'.format(col.name)
            assert col.server_default, 'Column "{}" should have server_default set'.format(col.name)

    def test_column_date_is_callable(self):
        for col_check in self.column_check_generator():
            if col_check.skip_callable_date_check:
                continue
            col = getattr(self.entity_cls, col_check.name)
            if not isinstance(col.type, (sa.Date, sa.DateTime, ArrowType)):
                continue
            datetypes = (datetime.date, datetime.datetime, arrow.Arrow)
            if col.default and isinstance(col.default.arg, datetypes):
                raise AssertionError(f'Column {col.name} default value is set to an instantiated'
                                     f' date type, this is probably not what you want.')
            if col.onupdate and isinstance(col.onupdate.arg, datetypes):
                raise AssertionError(f'Column {col.name} onupdate value is set to an instantiated'
                                     f' date type, this is probably not what you want.')

    def test_column_numeric_scale_precision_set(self):
        for col_check in self.column_check_generator():
            col = getattr(self.entity_cls, col_check.name)
            if isinstance(col.type, sa.Numeric) and not isinstance(col.type, sa.Float):
                assert col.type.precision is not None, \
                    'Column "{}" does not specify precision'.format(col.name)
                assert col.type.scale is not None, \
                    'Column "{}" does not specify scale'.format(col.name)

    def test_all_columns_are_constraint_tested(self):
        """Checks that all fields declared on entity are in the constraint tests"""

        expected_columns = [col.name for col in self.entity_cls.__table__.columns]
        constraint_columns = [col[0] for col in self.column_checks]
        inherited_columns = []

        # Even if the entity class is based (eventually) on DefaultColsMixin, polymorphic subclasses
        #   will not have the default cols, as the parent entity alone will have them
        if isinstance(self.entity_cls(), DefaultColsMixin) and not \
                getattr(self.entity_cls, '__mapper_args__', {}).get('polymorphic_identity', None):
            # Include columns from common import base classes
            for field in vars(DefaultColsMixin):

                # Ignore the usual __class__, __dict__, blah, blah blah
                if field.startswith('__'):
                    continue

                col = getattr(DefaultColsMixin, field)

                # Only consider SQLAlchemy columns
                if isinstance(col, sa.sql.schema.Column):
                    inherited_columns.append(field)

        combined_columns = set(constraint_columns + inherited_columns)
        if len(combined_columns) != len(expected_columns):

            missing = set(expected_columns) - combined_columns

            raise AssertionError(
                'Missing {} constraint tests for {}.'.format(
                    self.entity_cls.__name__,
                    ', '.join(missing))
            )

    def check_unique_constraint(self, **kwargs):
        self.entity_cls.fake(**kwargs)
        try:
            self.entity_cls.fake(**kwargs)
            raise AssertionError('Uniqueness error was not encountered.')
        except Exception as e:
            if not validate_unique_exc(e):
                raise


class FormBase(object):
    """Base class for testing forms. Assumes WTForms framework.

    Class attributes:

        form_cls: Form class to verify
        form_cls_factory: Method returning a form class (use instead of form_cls)
    """
    form_cls = None
    form_cls_factory = None

    def ok_data(self, **kwargs):
        """Returns data that should pass validation by default.

        Provided kwargs should override the default set of data.
        """
        return kwargs

    def test_disallowed_fields_not_present(self):
        """Most forms will not want to have the record timestamps as fields. Check that here."""
        form = self.create_form()
        assert "created_utc" not in form
        assert "updated_utc" not in form

    def test_ok_data_valid(self):
        """Assert that ``ok_data`` has valid data by default."""
        self.assert_valid()

    def create_form(self, obj=None, **kwargs):
        """Creates a CSRF-less form instance.

        Form data comes from ``ok_data`` with kwargs applied, or from a ``_form_data`` kwarg if
        that is provided (i.e. bypasses ``ok_data`` defaults).

        Object may be provided with the ``obj`` kwarg. Default None.
        """
        form_cls = kwargs.pop('form_cls', None) or self.form_cls
        if not form_cls and self.form_cls_factory:
            form_cls = self.form_cls_factory()

        data = kwargs.pop('_form_data', self.ok_data(**kwargs))
        # attempt to disable CSRF here with a patch. This covers a majority of cases in testing
        with mock.patch.dict(current_app.config, WTF_CSRF_ENABLED=False):
            form = form_cls(MultiDict(data), obj=obj)

        return form

    def assert_invalid(self, **kwargs):
        """Create a form with the given kwargs, assert it is invalid, and return the form."""
        form = self.create_form(**kwargs)
        assert not form.validate(), "expected form errors"
        return form

    def assert_valid(self, **kwargs):
        """Create a form with the given kwargs, assert it is valid, and return the form."""
        form = self.create_form(**kwargs)
        assert form.validate(), form.errors
        return form

    def verify_field(
        self, name, label=None, required=False, choice_values=None, prefix=None, suffix=None
    ):
        """Create a form and verify a field's parameters are configured."""
        if required is not None:
            data = {name: ""}
            form = self.create_form(**data)
            field = form[name]
            form.validate()

            if required:
                assert "This field is required." in field.errors
            else:
                assert not field.errors, field.errors
        else:
            form = self.create_form()
            field = form[name]

        if hasattr(field, "choices"):
            if choice_values:
                assert choice_values == [choice[0] for choice in field.choices]

        if label:
            assert field.label.text == label

        if prefix or suffix:
            form = self.create_form()
            field = form[name]

            pq = PyQuery(field())
            if prefix:
                assert pq(".input-group-prepend").text() == prefix
            if suffix:
                assert pq(".input-group-append").text() == suffix



import datetime as dt
import operator
import random

import arrow
import blazeutils.strings
import pytz
import sqlalchemy as sa
import wrapt
from blazeutils import tolist
from keg.db import db
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy_utils import ArrowType, EmailType

import keg_elements.db.columns as columns
import keg_elements.db.utils as dbutils
import keg_elements.decorators as decor
from keg_elements.extensions import lazy_gettext as _

might_commit = decor.keyword_optional('_commit', after=dbutils.session_commit, when_missing=True)
"""Decorator directing the wrapped method to commit db session upon completion.

A `_commit` bool kwarg is added to the wrapped method's definition, allowing a developer to
turn off the commit when calling.

Exceptions during commit are raised after the session is rolled back.
"""
might_flush = decor.keyword_optional('_flush', after=dbutils.session_flush)
"""Decorator directing the wrapped method to flush db session upon completion.

A `_flush` bool kwarg is added to the wrapped method's definition, allowing a developer to
turn off the flush when calling.

Exceptions during flush are raised after the session is rolled back.
"""


@wrapt.decorator
def kwargs_match_entity(wrapped, instance, args, kwargs):
    """
    Asserts that the kwargs passed to the wrapped method match the columns/relationships
    of the entity.
    """
    if kwargs.get('_check_kwargs', True):
        insp = sa.inspection.inspect(instance)

        # Only allow kwargs that correspond to a column or relationship on the entity
        allowed_keys = {col.key for col in insp.columns} | set(insp.relationships.keys())

        # Ignore kwargs starting with "_"
        kwarg_keys = set(key for key in kwargs if not key.startswith('_'))
        extra_kwargs = kwarg_keys - allowed_keys
        assert not extra_kwargs, _('Unknown column or relationship names in kwargs: {kwargs!r}',
                                   kwargs=sorted(extra_kwargs))
    return wrapped(*args, **kwargs)


class DefaultColsMixin(object):
    """Basic entity mixin providing int primary key and created/updated timestamps."""
    id = sa.Column('id', sa.Integer, primary_key=True)
    created_utc = sa.Column(ArrowType, nullable=False, default=arrow.utcnow,
                            server_default=dbutils.utcnow())
    updated_utc = sa.Column(ArrowType, nullable=False, default=arrow.utcnow, onupdate=arrow.utcnow,
                            server_default=dbutils.utcnow())


class MethodsMixin:
    """Entity mixin providing developer/testing-centered methods."""
    def from_dict(self, data):
        """Update the instance with the passed information in `data`.

        `from_dict` will also update relationships either, 1:1, 1:N, M:M.

        .. note::
            `from_dict` intentionally does not commit and for related entities
            turns off auto flushing. This is to prevent premature flushes with
            incomplete objects
        """

        def update_related(key, value, prop):
            """Update the entity based on the type of relationship it is"""
            related_class = prop.mapper.class_

            with db.session.no_autoflush:
                if prop.uselist:
                    new_list = list()
                    for row in value:
                        obj = related_class.add_or_edit(row, _commit=False)
                        new_list.append(obj)
                    setattr(self, key, new_list)
                else:
                    setattr(self, key, related_class.add_or_edit(value, _commit=False))

        mapper = self.__mapper__
        entity_props = {attr.key: attr for attr in mapper.attrs}

        for key, value in data.items():
            prop = entity_props.get(key)
            is_column = isinstance(prop, sa.orm.properties.ColumnProperty)
            is_relationship = isinstance(prop, sa.orm.properties.RelationshipProperty)

            if prop is None:
                continue
            elif is_column:
                setattr(self, key, value)
            elif is_relationship:
                update_related(key, value, prop)
            else:
                raise NotImplementedError(
                    _('Updating property types is not implemented.'))

    def to_dict(self, exclude=frozenset(), hybrids=frozenset()):
        """Covert the object properties to a dictionary.

        :param exclude: a list of columns to ignore
        :param hybrids: a list of the hybrid properties to include in the dictionary.
        :returns: a dictionary representation of the object

        .. note: By default hybrid properties are not included in the returned dict. To add a hybrid
            property to the returned dict pass a list of the property names and they will be
            included.
        """
        data = dict((name, getattr(self, name))
                    for name in self.column_names()
                    if name not in exclude)

        for hybrid in hybrids:
            data[hybrid] = getattr(self, hybrid)

        return data

    @classmethod
    def column_names(cls):
        """Return a set of column keys, which may not match attribute names."""
        return {col.key for col in cls.__mapper__.columns}

    @classmethod
    def primary_keys(cls):
        """Helper to get the table's primary key columns."""
        return cls.__table__.primary_key.columns

    @might_commit
    @might_flush
    @kwargs_match_entity
    @classmethod
    def add(cls, **kwargs):
        """Create a new persisted record constructed from the given kwargs.

        :param _commit: enable/disable commit. Default True.
        :param _flush: enable/disable flush. Default True.
        :return: entity instance created and optionally persisted.
        """
        obj = cls()
        obj.from_dict(kwargs)
        db.session.add(obj)
        return obj

    @might_commit
    @might_flush
    @classmethod
    def delete(cls, oid):
        """Delete an object from the session

        :param oid: the object identifier, normally the primary key
        :rtype: bool
        :return: The result of the operation
        """
        obj = cls.get(oid)

        if obj is None:
            return False

        db.session.delete(obj)
        return True

    @might_commit
    @classmethod
    def delete_cascaded(cls):
        """For testing, remove all records from the table. Extend for dependencies.

        By default, this affects only this entity. By design, though, the entity should
        override to call `delete_cascaded` on any entities that have foreign key
        dependence on this entity. Key cascades may cover some of these cases, but
        db cascades are not always desireable, and tests often need to easily clear
        a number of tables to ensure good starting state.
        """
        cls.query.delete(synchronize_session=False)
        db.session.expire_all()

    @might_commit
    @might_flush
    @kwargs_match_entity
    @classmethod
    def edit(cls, _oid=None, **kwargs):
        """Edit an object in session with the kwargs, and optionally flush or commit.

        :param oid: the object identifier, normally the primary key
        :param _commit: enable/disable commit. Default True.
        :param _flush: enable/disable flush. Default True.
        :return: entity instance edited and optionally flushed/committed
        """
        try:
            primary_keys = _oid or [kwargs.get(x.name)
                                    for x in cls.primary_keys()
                                    if x is not None and kwargs.get(x.name) is not None]
            if not primary_keys:
                raise KeyError
        except KeyError:
            raise AttributeError(_('No primary key was found in `oid` or `kwargs`'
                                   ' for which to retrieve the object to edit'))

        obj = cls.get(primary_keys)
        obj.from_dict(kwargs)
        return obj

    @classmethod
    def insert(cls, values=None, **kwargs):
        """Similar to ``add`` but without the ORM overhead. Useful for high data throughput
        cases where having kwargs name validation and ORM ops/session on every iteration
        would be inefficient.

        Assumes the calling code is handling session flush/commit.

        :param values: optional dict of values to insert
        :param kwargs: values to insert, can be combined with the ``values`` dict
        :return: primary key value(s). Note: SQLite does not support this
        """
        if values is None:
            values = kwargs
        else:
            values.update(kwargs)

        stmt = sa.insert(cls.__table__).values(**values)
        if db.engine.dialect.name == 'sqlite':
            return db.session.execute(stmt)
        primary_keys = cls.primary_keys()
        stmt = stmt.returning(*primary_keys)
        result = db.session.execute(stmt)
        if len(primary_keys) > 1:
            return result.fetchone()
        return result.scalar()

    @classmethod
    def update(cls, ent_id, values=None, **kwargs):
        """Similar to ``edit`` but without the ORM overhead. Useful for high data throughput
        cases where having kwargs name validation and ORM ops/session on every iteration
        would be inefficient.

        Assumes the calling code is handling session flush/commit.

        Note: if the entitiy has multiple primary key columns, ``ent_id`` should be an iterable
        with the values to match in the order of the column definitions as specified in the
        SA model (i.e. the order of columns returned by ``cls.primary_keys()``).

        :param ent_id: primary key value(s) to match for update
        :param values: optional dict of values to insert
        :param kwargs: values to insert, can be combined with the ``values`` dict
        :return: db cursor result
        """
        if values is None:
            values = kwargs
        else:
            values.update(kwargs)

        stmt = (
            sa.update(cls.__table__)
            .values(**values)
            .where(
                *map(
                    lambda pair: pair[0] == pair[1],
                    zip(cls.primary_keys(), tolist(ent_id))
                )
            )
        )
        return db.session.execute(stmt)

    @classmethod
    def get(cls, ident, **kwargs):
        """Wraps db.session.get"""
        return db.session.get(cls, ident, **kwargs)

    @classmethod
    def get_by(cls, **kwargs):
        """Returns the instance of this class matching the given criteria or
        None if there is no record matching the criteria.

        If multiple records are returned, an exception is raised.
        """
        return cls.query.filter_by(**kwargs).one_or_none()

    @classmethod
    def get_where(cls, *clauses):
        """
        Returns the instance of this class matching the given clause(s) or None
        if there is no record matching the criteria.

        If multiple records are returned, an exception is raised.
        """
        return cls.query.filter(*clauses).one_or_none()

    @classmethod
    def pairs(cls, key_field, value_field, order_by=(), query=None,
              items=None):
        """Return a list of two item tuples

        :param key_field: string representing the key
        :param value_field: string representing the value
        :param order_by: iterable of columns to order the query by
        :param query: a base query from which to generate the pairs
        :param items: a function which takes one record returned by query and
                      returns the tuple object
        """

        items = items if items else operator.attrgetter(key_field, value_field)
        query = query or cls.query
        result = query.order_by(*order_by).all()

        return [items(obj) for obj in result]

    @kwargs_match_entity
    @classmethod
    def fake(cls, **kwargs):
        """Create an object for testing with default data appropriate for the field type

        * Will automatically set most field types ignoring those passed in via kwargs.
        * Subclasses that have foreign key relationships should setup those relationships before
          calling this method. See `testing_set_related` for additional information.

        Random data that is set on a column comes from one of these sources:

        * `random_data_for_column` entity method provides randoms for most normal column types
        * `randomdata` is given in column info as the name of an entity method to call for data::

            class MyEntity(MethodsMixin, db.Model):
                foo = sa.Column(sa.Unicode, info={'randomdata': 'foo_generator'})

                @classmethod
                def foo_generator(cls):
                    return 'bar'

        * `random_magnitude` is given in column info to be treated as the +/- random range.
        * `random_range` is given specifically as a low/high random range.

        Special kwargs:
        _numeric_defaults_range: a tuple of (HIGH, LOW) which controls the acceptable defaults of
        the two number types. Both integer and numeric (float) fields are controlled by
        this setting.
        """

        numeric_range = kwargs.pop('_numeric_defaults_range', None)

        insp = sa.inspection.inspect(cls)

        skippable = lambda column: (
            column.key in kwargs                # skip fields already in kwargs
            or cls.testing_skip_column(column)  # skip fields defined by class rules
        )

        for column in (col for col in insp.columns if not skippable(col)):
            try:
                kwargs[column.key] = cls.random_data_for_column(
                    column, numeric_range)
            except ValueError:
                pass

        return cls.add(**kwargs)

    @classmethod
    def testing_skip_column(cls, column):
        is_property = lambda column: isinstance(column, sa.sql.elements.Label)
        return (
            is_property(column)       # skip column properties
            or column.foreign_keys    # skip foreign keys
            or column.server_default  # skip fields with server defaults
            or column.default         # skip fields with defaults
            or column.primary_key     # skip any primary key
        )

    @classmethod
    def random_data_for_column(cls, column, numeric_range):  # noqa: C901
        """Provides random testing data for a number of column types.

        Raises a ValueError if the type is not handled. In that case, override as needed.
        """
        if 'randomdata' in column.info:
            if type(column.info['randomdata']) is str:
                # assume randomdata the is name of a method on the class
                callable = getattr(cls, column.info['randomdata'])
                data = callable()
                return data

            return column.info['randomdata']()

        default_range = (-100, 100) if numeric_range is None else numeric_range
        if isinstance(column.type, sa.types.Enum):
            return random.choice(column.type.enums)
        elif isinstance(column.type, sa.types.Boolean):
            return random.choice([True, False])
        elif isinstance(column.type, sa.types.Integer):
            if numeric_range is not None:
                return random.randint(*default_range)
            return dbutils.random_int(column, default_range)
        elif isinstance(column.type, sa.types.Float):
            return random.uniform(*default_range)
        elif isinstance(column.type, sa.types.Numeric):
            if numeric_range is not None or column.type.scale is None:
                return random.uniform(*default_range)
            return dbutils.random_numeric(column)
        elif isinstance(column.type, sa.types.Date):
            return dt.date.today()
        elif isinstance(column.type, sa.types.DateTime):
            if hasattr(dt, 'UTC'):
                # Return a TZ-aware datetime if possible, since python is deprecating
                # the utcnow method
                return dt.datetime.now(dt.UTC)
            return dt.datetime.utcnow()
        elif isinstance(column.type, ArrowType):
            return arrow.utcnow()
        elif isinstance(column.type, EmailType):
            return dbutils.randemail(min(column.type.length or 50, 50))
        elif isinstance(column.type, columns.TimeZoneType):
            return random.choice(pytz.common_timezones)
        elif isinstance(column.type, (sa.types.String, sa.types.Unicode)):
            return blazeutils.strings.randchars(min(column.type.length or 25, 25))
        raise ValueError(_('No randomization for this column type'))

    @classmethod
    def testing_set_related(cls, kwargs, model,
                            *testing_create_args, **testing_create_kwargs):
        """Create a related object for testing, if it is not specified in kwargs.

        Designed to be used by `testing_create`. A common issue is that related
        test records need to be set up with the test entity instance, but they
        could be specified already in kwargs. In addition, relationships already
        specified may be given on the relationship attribute or the foreign-key
        field.

        This method takes existing `testing_create` kwargs and the related
        entity, and makes the necessary updates.

        Relationship name is generated from the given entity by default, but may
        be passed in as a `_relationship_name` keyword argument. The foreign key
        field is assumed to be the relationship name with an `_id` suffix, but
        may be specified with the `_relationship_field` keyword argument.

        Any additional args/kwargs are passed to the given model's `testing_create`.
        """
        relationship_name = testing_create_kwargs.pop(
            '_relationship_name',
            blazeutils.strings.case_cw2us(model.__name__),
        )
        relationship_field = testing_create_kwargs.pop(
            '_relationship_field',
            '{}_id'.format(relationship_name),
        )

        for attr_name in (relationship_name, relationship_field):
            if not hasattr(cls, attr_name):
                print(cls, attr_name)
                raise Exception(
                    'testing_set_related expects "{}" on class {}'.format(
                        attr_name,
                        cls.__name__,
                    )
                )

        # generic logic to respect existing kwargs setting on either field
        if not {relationship_name, relationship_field} & set(kwargs.keys()):
            kwargs[relationship_name] = model.fake(
                *testing_create_args,
                **testing_create_kwargs
            )

    @might_commit
    @might_flush
    @classmethod
    def add_or_edit(cls, data):
        """Creates or updates the record associated with `data`

        `add_or_edit` supports multiple primary key entities.
        """
        if isinstance(data, db.Model) or not data:
            return data

        primary_keys = [
            data.get(x.name) for x in cls.primary_keys()
            if x is not None and data.get(x.name) is not None
        ]
        obj = cls.get(primary_keys) if primary_keys else None

        return (cls.add(_commit=False, **data)
                if obj is None
                else cls.edit(primary_keys, _commit=False, **data))

    def update_collection(self, attr_name, data):
        """Update the specified relationship collection with the given data.

        Attempt will be made to match existing collection records to update. If an existing
        record is not in the data set, it will be removed.

        :param attr_name: relationship attribute to update.
        :param data: iterable of dicts representing records to add/update.
        """
        dbutils.CollectionUpdater(self, attr_name, data).update()


class DefaultMixin(DefaultColsMixin, MethodsMixin):
    """Entity mixin combining DefaultColsMixin and MethodsMixin."""
    pass


class HardDeleteProhibitedError(Exception):
    pass


class SoftDeleteMixin:
    """SoftDeleteMixin alters the way deletes are performed by adding a deleted_utc column

    A soft-delete doesn't actually remove the row, instead, it adds a time to the ``deleted_utc``
    column which indicates when it was deleted.

    .. note (NZ): This can complicate unique constraints, joins, and general business logic, but is
        rather useful when you can't delete an object outright because it is connected to other
        permanent objects which should never be deleted.

    .. note (NZ): ``SoftDeleteMixin`` should appear before ``MethodsMixin`` as the parent class of
        the created entity so that the ``delete`` and ``testing_create`` methods are called in the
        correct order.

        class MyTable(SoftDeleteMixin, MethodsMixin, Model):
            column1 = sa.Column(sa.Numeric)


    .. note (NZ): ``testing_create`` takes a special ``_is_deleted`` flag which enables you to
        create an already deleted record or you can pass ``deleted_utc`` manually.

    .. note (NZ): There is a new event registered within SQLAlchemy to prevent accidental deletions
        of entities that inherit from this class.

    This event is not fired for bulk operations (``session.delete``) and makes it possible to delete
    everything.

    To have finer grained control of this event for every entity, you can override the behavior by
    implementing ``before_delete_event`` on the entity::

        class MyEntity(SoftDeleteMixin, Model):
            column1 = sa.Column(sa.Numeric)

            # `mapper` and `connection` are passed through from the parent handler
            def before_delete_event(self, mapper, connection):
                # .. custom logic
    """

    deleted_utc = sa.Column(ArrowType, nullable=True)

    @might_commit
    @might_flush
    @classmethod
    def delete(cls, oid):
        """Add the deleted_utc timestamp

        :param oid: the object identifier, normally the primary key
        :rtype: bool
        :return: The result of the operation
        """
        return cls.query.filter(cls.id == oid).update(
            {"deleted_utc": arrow.utcnow()},
            synchronize_session=False
        ) > 0

    @might_commit
    @might_flush
    @classmethod
    def fake(cls, *args, _is_deleted=False, **kwargs):
        kwargs.setdefault('deleted_utc', arrow.utcnow() if _is_deleted else None)
        return super().fake(*args, **kwargs)

    @staticmethod
    def sqla_before_delete_event(mapper, connection, target):
        if hasattr(target, 'before_delete_event'):
            target.before_delete_event(mapper, connection)
        else:
            raise HardDeleteProhibitedError(
                'Unable to completely delete {}, this object implements soft-deletes.'.format(target))  # noqa

    @might_commit
    @classmethod
    def delete_cascaded(cls):
        cls.query.delete(synchronize_session=False)


sa.event.listen(SoftDeleteMixin, 'before_delete', SoftDeleteMixin.sqla_before_delete_event,
                propagate=True)


class LookupMixin(SoftDeleteMixin):
    """Provides a base for id/label pair tables, used in one-to-many relationships.

    Based on SoftDeleteMixin, so any lookup record that is deleted/deactivated is
    still available for existing records.

    A code field is provided for developer reference in code, so a changeable label
    does not need to be hard-coded for lookup.

    Developer expectations:
    - LookupMixin will precede DefaultMixin or MethodsMixin in entity base classes
    - Only active labels will be listed for linking to new related records
    - `include_ids` will be used for ensuring existing records preserve lookup
    """

    label = sa.Column(sa.Unicode(255), nullable=False, unique=True)
    code = sa.Column(sa.Unicode(255))

    @hybrid_property
    def is_active(self):
        """Hybrid property returning a record's active status.

        By default, deleted == inactive.
        """
        return self.deleted_utc is not None

    @is_active.expression
    def is_active(cls):
        return sa.sql.case((cls.deleted_utc.is_(None), sa.true()), else_=sa.false())

    @classmethod
    def _active_query(cls, include_ids=None, order_by=None):
        if order_by is None:
            order_by = cls.label

        if include_ids:
            include_ids = tolist(include_ids)
            clause = sa.sql.or_(
                cls.is_active == sa.true(),
                cls.id.in_(include_ids)
            )
        else:
            clause = cls.is_active == sa.true()

        return cls.query.filter(clause).order_by(order_by)

    @classmethod
    def list_active(cls, include_ids=None, order_by=None):
        """Fetch all records marked as active.

        :param include_ids: iterable of int ids that should be included even if inactive.
        :param order_by: column designation to use for SQLAlchemy when sorting values.
        """
        return cls._active_query(include_ids, order_by).all()

    @classmethod
    def pairs_active(cls, include_ids=None, order_by=None):
        """Fetch a list of (id, label) pairs for active records.

        :param include_ids: iterable of int ids that should be included even if inactive.
        :param order_by: column designation to use for SQLAlchemy when sorting values.
        """
        query = cls._active_query(include_ids, order_by)
        return cls.pairs('id', 'label', query=query)

    @classmethod
    def get_by_label(cls, label):
        """Fetch a lookup record by its label field."""
        return cls.get_by(label=label)

    @classmethod
    def get_by_code(cls, code):
        """Fetch a lookup record by its code (internal) field."""
        return cls.get_by(code=code)

    def __repr__(self):
        return '<{} {}:{}>'.format(self.__class__.__name__, self.id, self.label)



from webgrid import Column
from webgrid.filters import TextFilter

from kegel_app.extensions import Grid
from kegel_app.model import entities as ents


class DemoGrid(Grid):
    Column('Name', ents.Thing.name, TextFilter)



from keg.testing import ContextManager
import pytest

from kegel_app.app import KegElApp


def pytest_configure(config):
    app = KegElApp.testing_prep()

    from keg.db import db
    with app.app_context():
        print('Running tests against {} DB'.format(db.engine.dialect.name))


@pytest.fixture(scope='class', autouse=True)
def auto_app_context():
    with ContextManager.get_for(KegElApp).app.app_context():
        yield



# coding=utf-8
import markupsafe
import pytest
import wtforms
from werkzeug.datastructures import MultiDict

import keg_elements.forms.validators as validators
from keg_elements.forms import ModelForm


class UniqueForm(ModelForm):
    uq_field = wtforms.fields.StringField('thing', validators=[
        validators.ValidateUnique(object_html_link=lambda field: 'link')])

    def get_object_by_field(self, field):
        return 1 if field.data == '1' else None


class TestUniqueValidator(object):
    def test_validation_passes(self):
        form = UniqueForm(MultiDict({'uq_field': '1'}), obj=1)
        assert form.uq_field.validate(form) is True

        form = UniqueForm(MultiDict({'uq_field': '2'}), obj=1)
        assert form.uq_field.validate(form) is True

    def test_validation_fails(self):
        form = UniqueForm(MultiDict({'uq_field': '1'}))
        assert form.uq_field.validate(form) is False
        assert form.uq_field.errors == [
            markupsafe.Markup('This value must be unique but is already assigned to link.')
        ]

        form = UniqueForm(MultiDict({'uq_field': '1'}), obj=2)
        assert form.uq_field.validate(form) is False
        assert form.uq_field.errors == [
            markupsafe.Markup('This value must be unique but is already assigned to link.')
        ]

    def test_no_object_link_provided(self):
        class Form(ModelForm):
            uq_field = wtforms.fields.StringField('thing', validators=[validators.ValidateUnique()])

            def get_object_by_field(self, field):
                return 1 if field.data == '1' else None

        form = Form(MultiDict({'uq_field': '1'}))
        assert form.uq_field.validate(form) is False
        assert form.uq_field.errors == [
            markupsafe.Markup('This value must be unique but is already assigned.')
        ]

    def test_get_obj(self):
        class Form:
            def __init__(self, **kwargs):
                self.__dict__.update(kwargs)

        validator = validators.ValidateUnique()

        assert validator.get_obj(Form(obj='foo')) == 'foo'
        assert validator.get_obj(Form(_obj='foo')) == 'foo'

        with pytest.raises(AttributeError) as exc:
            validator.get_obj(Form())

        assert str(exc.value) == (
            'Form must provide either `obj` or `_obj` property for uniqueness validation.'
        )


class TestAlphaNumericValidator(object):

    def test_alphanumeric(self):
        message = 'Must only contain alphanumeric data.'

        class AlphaNumericMockForm(wtforms.Form):
            alpha_numeric_field = wtforms.StringField('AlphaNumeric', validators=[
                validators.ValidateAlphaNumeric()
            ])

        form = AlphaNumericMockForm(alpha_numeric_field='123456asb')
        form.validate()
        assert form.errors == {}

        form = AlphaNumericMockForm(alpha_numeric_field='123456')
        form.validate()
        assert form.errors == {}

        form = AlphaNumericMockForm(alpha_numeric_field='abcd')
        form.validate()
        assert form.errors == {}

        form = AlphaNumericMockForm(alpha_numeric_field='123456!')
        form.validate()
        assert form.errors['alpha_numeric_field'] == [message]

        form = AlphaNumericMockForm(alpha_numeric_field='123456!')
        form.validate()
        assert form.errors['alpha_numeric_field'] == [message]

        form = AlphaNumericMockForm(alpha_numeric_field='!212')
        form.validate()
        assert form.errors['alpha_numeric_field'] == [message]

        form = AlphaNumericMockForm(alpha_numeric_field=' ')
        form.validate()
        assert form.errors['alpha_numeric_field'] == [message]

        form = AlphaNumericMockForm(alpha_numeric_field='123dsaf4 ')
        form.validate()
        assert form.errors['alpha_numeric_field'] == [message]

        form = AlphaNumericMockForm(alpha_numeric_field=' 123afd4')
        form.validate()
        assert form.errors['alpha_numeric_field'] == [message]

        form = AlphaNumericMockForm(alpha_numeric_field='1 23fdas4')
        form.validate()
        assert form.errors['alpha_numeric_field'] == [message]

        form = AlphaNumericMockForm(alpha_numeric_field=u'åfasdf')
        form.validate()
        assert form.errors['alpha_numeric_field'] == [message]



from decimal import Decimal

import markupsafe
from wtforms import ValidationError
import re

from keg_elements.extensions import lazy_gettext as _


class ValidateAlphaNumeric(object):
    """
        A validator to make sure than a form field contains only alphanumeric data

        Example::
            import keg_elements.forms.validators as validators

            wtforms.StringField('AlphaNumeric', validators=[validators.ValidateAlphaNumeric()])
    """
    regex = re.compile(r'^[a-zA-Z0-9]+$')

    def __init__(self, message=None):
        self.message = message

    def __call__(self, form, field):
        value = field.data

        message = self.message
        if message is None:
            message = field.gettext(_("Must only contain alphanumeric data."))

        if not self.regex.match(value):
            raise ValidationError(message)


def numeric(form, field):
    try:
        int(field.data)
    except ValueError:
        raise ValidationError(_('Value is not numeric.'))


class NumberScale(object):
    """Ensure the field's numeric value is within a maximum number of decimal places."""
    def __init__(self, scale, message=None):
        self.scale = scale
        if not message:
            message = _(u'Field must have no more than {scale} decimal places.', scale=scale)
        self.message = message

    def __call__(self, form, field):
        # if the data is not decimal, don't do anything with it (could be an alpha string)
        if not isinstance(field.data, Decimal):
            return
        # use decimal's quantization to see if it's equal to field data at the full scale.
        #   If it isn't, user entered too many decimal places
        if field.data != field.data.quantize(Decimal(1) / 10 ** self.scale):
            raise ValidationError(self.message)


class ValidateUnique(object):
    """Ensure the field's value is unique from other values

    This validator requires a ``get_object_by_field(field)`` on the form which
    returns an object (if one exists) or None to check for uniqueness.

    Pass ``object_html_link`` to this validator's ``__init__`` to get an error
    message which contains a hyperlink to the existing object.

    .. note :: Combine with ``wtforms.validators.optional`` when a field is nullable to
        allow for None values.
    """
    field_flags = {'unique': True}

    def __init__(self, object_html_link=None):
        self.object_html_link = object_html_link

    def get_obj(self, form):
        if hasattr(form, 'obj'):
            return form.obj
        if hasattr(form, '_obj'):
            return form._obj

        raise AttributeError(_(
            'Form must provide either `obj` or `_obj` property for uniqueness validation.'
        ))

    def __call__(self, form, field):
        obj = self.get_obj(form)
        other = form.get_object_by_field(field)

        both_exist = None not in (obj, other)
        same_record = both_exist and other == obj
        another_exists_with_value = obj is None and other  # new obj with existing object with value

        if (both_exist and not same_record) or another_exists_with_value:
            if self.object_html_link is None:
                text = _('This value must be unique but is already assigned.')
            else:
                text = _('This value must be unique but is already assigned to {link}.',
                         link=self.object_html_link(other))
            msg = markupsafe.Markup(text)
            raise ValidationError(msg)
        return True



import os


class DefaultProfile(object):
    KEG_KEYRING_ENABLE = False

    SQLALCHEMY_TRACK_MODIFICATIONS = True
    SQLALCHEMY_DATABASE_URI = os.environ.get('SQLALCHEMY_DATABASE_URI', 'sqlite://')


class TestProfile(object):
    KEG_KEYRING_ENABLE = False

    # Turn off CSRF protection for tests.  Makes form testing easier and provides no value during
    # testing.
    WTF_CSRF_ENABLED = False
    SQLALCHEMY_TRACK_MODIFICATIONS = True
    # SQLALCHEMY_DATABASE_URI = 'postgresql://postgres:@localhost/keg-elem-test'

    SQLALCHEMY_DATABASE_URI = os.environ.get('SQLALCHEMY_DATABASE_URI', 'sqlite://')



import copy

import sqlalchemy as sa


def add_non_nullable_column(op, table_name, column, default_value=None, default_value_query=None,
                            schema=None):
    """
    Create a new non-nullable column

    :param op: The operational module or batch operator imported from alembic.op
    :param table_name: The name of the table to add the column to
    :param column: An SQLAlchemy column declaration (i.e. `sa.Column()` expression)
    :param default_value: The initial value for the column.
        Cannot be used with the `default_value_query`
    :param default_value_query: An update query that sets the column value for all rows.
        Cannot be used with `default_value`
    :param schema: The name of the table's schema
    :return:
    """
    if default_value is None and default_value_query is None:
        raise ValueError('Must provide either default_value or default_value_query')
    if default_value is not None and default_value_query is not None:
        raise ValueError('You may not pass both default_value and default_value_query')

    column = copy.copy(column)
    column.nullable = True

    op.add_column(table_name, column, schema=schema)

    if default_value_query is not None:
        op.execute(default_value_query)
    else:
        op.execute(
            sa.Table(
                table_name,
                sa.MetaData(),
                sa.Column(column.name, column.type),
                schema=schema
            ).update().values({column.name: default_value})
        )

    op.alter_column(
        table_name,
        column.name,
        schema=schema,
        nullable=False,
        existing_type=column.type
    )


def postgres_update_enum_options(op, table_column_list, enum_name, new_values):
    """
    Update an enum's options within the migration transaction. In Postgres updating an enum is
    typically not permitted within a transaction. However, dropping and re-creating the type works.

    Note: this function will remove the default value on every column it updates so you will need
    to reset any server-side defaults after running this function.

    :param op: The operational module or batch operator imported from alembic.op
    :param table_column_list: A list of three item tuples (schema, table_name, column_name) for
        each column where this enum is used
    :param enum_name: The name of the enum type
    :param new_values: A list of the enum's updated values
    """
    old_type = sa.Enum(name=enum_name)
    new_type = sa.Enum(*new_values, name=enum_name)

    for schema, table_name, column_name in table_column_list:
        op.alter_column(
            table_name,
            column_name,
            schema=schema,
            type_=sa.Unicode,
            postgresql_using=f'{column_name}::text',
            server_default=None
        )

    old_type.drop(op.get_bind(), checkfirst=False)
    new_type.create(op.get_bind(), checkfirst=False)

    for schema, table_name, column_name in table_column_list:
        op.alter_column(
            table_name,
            column_name,
            schema=schema,
            type_=new_type,
            postgresql_using=f'{column_name}::{enum_name}'
        )


def postgres_get_enum_values(op, enum_name, schema=None):
    """
    Get a list of values for an existing enum type.

    :param op: The operational module or batch operator imported from alembic.op
    :param enum_name: The name of the enum type
    :return: A list of the enum's values
    """

    if schema:
        enum_name = f'{schema}.{enum_name}'

    results = op.get_bind().execute(
        sa.select(
            sa.func.unnest(
                sa.func.enum_range(
                    sa.text('NULL::{}'.format(enum_name))
                )
            )
        )
    )
    return [x for x, in results.fetchall()]



import functools
import inspect
import logging
import warnings
from decimal import Decimal
from operator import attrgetter

from blazeutils.strings import case_cw2dash
import flask
from flask_wtf import FlaskForm as BaseForm
from keg.db import db
import sqlalchemy as sa
from markupsafe import Markup
from sqlalchemy_utils import ArrowType, get_class_by_table
import wtforms.fields
import wtforms.form
from wtforms.validators import InputRequired, Optional, StopValidation, NumberRange, AnyOf
from wtforms_alchemy import (
    FormGenerator as FormGeneratorBase,
    model_form_factory,
    model_form_meta_factory,
)
from wtforms_components.fields import (
    SelectField as SelectFieldBase,
    SelectMultipleField as SelectMultipleFieldBase,
)

from keg_elements.db.columns import DBEnum
from keg_elements.db.utils import has_column
from keg_elements.extensions import lazy_gettext as _
from keg_elements.forms.validators import NumberScale

form_element = flask.Blueprint('form_element', __name__)
log = logging.getLogger(__name__)


def to_title_case(x):
    """ underscore or dash to title case notation """
    return x.replace('_', ' ').replace('-', ' ').title()


# sentinel
_not_given = ()


class FieldMeta(object):
    """Meta information for fields to override model-generated info.

    Rather than forcing all-or-nothing acceptance of model-generated meta info from wtforms,
    FieldMeta may be provided in the FieldsMeta nested class of the form to override specifics.

    All modifications are applied to the field instance during the form generation process.

    Example::

        class PersonForm(ModelForm):
            class Meta:
                model = Person

            class FieldsMeta:
                name = FieldMeta('Full Name')

    :param label_text: Force a label value.
    :param description: Force a description value.
    :param label_modifier: Callable to be called with the default label. label_text takes
        precedence if both are provided. But, this modifier will apply to choices as well if
        applicable and a choices_modifier is not given.
    :param choices_modifier: Callable to be called with the label value for choices.
    :param required: Force validators to be added/removed for requirement.
    :param widget: Force a specific widget to be used for the field in render.
    :param extra_validators: Add the given validators to those existing on the field.
    :param coerce: Applies a specific coerce for field values. Applicable to select fields only.
    :param default: Forces a default value on the field.
    """

    def __init__(self, label_text=_not_given, description=_not_given, label_modifier=_not_given,
                 choices_modifier=_not_given, choices=None, required=_not_given, widget=_not_given,
                 extra_validators=tuple(), coerce=_not_given, default=_not_given):
        self.label_text = label_text
        self.label_modifier = label_modifier
        self.description = description
        self.choices_modifier = choices_modifier
        self.choices = choices
        self.required = required
        self.widget = widget
        self.extra_validators = extra_validators
        self.coerce = coerce
        self.default = default

        assert self.required in (_not_given, False, True)

    def apply_to_field(self, field):
        # field is a wtforms.fields.core.UnboundField instance
        self.apply_to_label(field)
        self.apply_to_description(field)
        self.apply_to_choices(field)
        self.apply_required(field)
        self.apply_widget(field)
        self.apply_extra_validators(field)
        self.apply_coerce(field)
        self.apply_default(field)

    def apply_to_label(self, field):
        default_label = field.kwargs['label']
        if self.label_text is not _not_given:
            label_text = self.label_text
        elif self.label_modifier is None:
            label_text = default_label
        elif self.label_modifier is _not_given:
            label_text = to_title_case(default_label)
        else:
            label_text = self.label_modifier(default_label)
        field.kwargs['label'] = self.modify_label(label_text)

    def modify_label(self, label_text):
        """ for subclasses to easily modify the final label text value """
        return label_text

    def apply_to_description(self, field):
        default_description = field.kwargs.get('description')
        if self.description is _not_given:
            description = default_description
        else:
            description = self.description
        field.kwargs['description'] = self.modify_description(description)

    def modify_description(self, description):
        """ for subclasses to easily modify the final label text value """
        return description

    def apply_to_choices(self, field):
        default_choices = field.kwargs.get('choices', None)
        if default_choices is None:
            # this isn't a field that has choices
            return

        if self.choices_modifier is None:
            modifier = None
        elif self.choices_modifier is not _not_given:
            modifier = self.choices_modifier
        elif self.label_modifier is None:
            # no choices modifier and the label modifier is explicit, so no label modifier
            modifier = None
        elif self.label_modifier is _not_given:
            # title case to labels by default
            modifier = to_title_case
        else:
            # a label modifier was given, use that since no choices modifier was given to override
            modifier = self.label_modifier

        if self.choices is not None:
            choices = self.choices
        elif modifier is None:
            choices = default_choices
        else:
            choices = [(v, modifier(l)) for v, l in default_choices]

        field.kwargs['choices'] = self.modify_choices(choices)

    def modify_choices(self, choices):
        return choices

    def apply_coerce(self, field):
        if self.coerce is _not_given:
            return
        if not issubclass(field.field_class, wtforms.SelectField):
            raise ValueError('`coerce` argument may only be used for select fields')
        field.kwargs['coerce'] = self.coerce

    def apply_default(self, field):
        if self.default is _not_given:
            return
        field.kwargs['default'] = self.default

    def apply_required(self, field):
        validators = field.kwargs.get('validators', [])

        if self.required == _not_given:
            # required value not given on FieldMeta, don't make any changes
            pass
        elif self.required:
            # If a required validator isn't present, we need to add one.
            req_val_test = lambda val: hasattr(val, 'field_flags') and 'required' in val.field_flags
            if not list(filter(req_val_test, validators)):
                validators.append(InputRequired())

            # If an optional validator is present, we need to remove it.
            not_opt_val_test = lambda val: not hasattr(val, 'field_flags') or \
                'optional' not in val.field_flags
            not_opt_validators = list(filter(not_opt_val_test, validators))
            field.kwargs['validators'] = not_opt_validators
        else:
            # If an optional validator isn't present, we need to add one.
            opt_val_test = lambda val: hasattr(val, 'field_flags') and 'optional' in val.field_flags
            if not list(filter(opt_val_test, validators)):
                validators.append(Optional())

            # If a required validator is present, we need to remove it.
            non_req_val_test = lambda val: not hasattr(val, 'field_flags') or \
                'required' not in val.field_flags
            not_req_validators = list(filter(non_req_val_test, validators))
            field.kwargs['validators'] = not_req_validators

    def apply_widget(self, field):
        if self.widget != _not_given:
            field.kwargs['widget'] = self.widget

    def apply_extra_validators(self, field):
        field.kwargs.setdefault('validators', [])
        field.kwargs['validators'] += self.extra_validators


def select_coerce(es_pass_thru, coerce, value):
    if es_pass_thru and value == '':
        return value
    if coerce is not _not_given:
        return coerce(value)

    # try coercing to int first.  If not valid, fall back to default behavior
    try:
        return int(value)
    except ValueError as e:
        if 'invalid literal for int()' not in str(e):
            raise
        return str(value)


class SelectMixin:
    def __init__(self, *args, **kwargs):
        self.add_blank_choice = kwargs.pop('add_blank_choice', True)
        coerce_arg = kwargs.pop('coerce', _not_given)
        super().__init__(*args, **kwargs)

        if self.add_blank_choice:
            # If we are adding a blank choice, and it is selected, we want the value that comes back
            # in .data to be None -> as if no value was selected.
            #
            # self.filters is a tuple, so have to do some extra work.
            self.filters = [lambda x: None if x == '' else x] + list(self.filters)

        self.coerce = functools.partial(select_coerce, self.add_blank_choice, coerce_arg)

    def iter_choices(self):
        if self.add_blank_choice:
            template_list = list(self._choices_generator([('', '')]))
            template = list(template_list[0])
            template[2] = (self.coerce, False)
            yield tuple(template)
        for value in super().iter_choices():
            yield value

    @property
    def choice_values(self):
        values = super().choice_values
        if self.add_blank_choice:
            return [''] + values
        return values

    @property
    def selected_choice_label(self):
        value_dict = dict(self.concrete_choices)
        return value_dict.get(self.data)


class SelectField(SelectMixin, SelectFieldBase):
    """
        Provides helpful features above wtforms_components SelectField which it is based on:

        1) Adds a blank choice by default at the front of the choices.  This results in your user
           being forced to select something if the field is required, which avoids initial
           defaulting of the first value in the field getting submitted.
        2) The coerce function used for the choices will automatically convert to int if possible,
           falling back to unicode if the value is not an integer.
    """


class SelectMultipleField(SelectMixin, SelectMultipleFieldBase):
    """
        Provides helpful features above wtforms_components SelectMultipleField which it is
        based on:

        The coerce function used for the choices will automatically convert to int if possible,
        falling back to unicode if the value is not an integer.
    """
    def __init__(self, *args, **kwargs):
        kwargs['add_blank_choice'] = kwargs.get('add_blank_choice', False)
        super().__init__(*args, **kwargs)


class MultiCheckboxField(wtforms.fields.SelectMultipleField):
    """
    A multiple-select, except displays a list of checkboxes.
    """


class RequiredBoolRadioField(wtforms.fields.RadioField):
    """A radio group field with true/false labels and a required validator.

    :param true_label: Optional, defaults to Yes.
    :param false_label: Optional, defaults to No.
    :param validators: Optional. Any provided validators will be added to InputRequired.
    """
    def __init__(self, *args, **kwargs):
        true_label = kwargs.pop('true_label', _('Yes'))
        false_label = kwargs.pop('false_label', _('No'))

        def bool_coerce(val):
            if val == u'True':
                return True
            if val == u'False':
                return False
            return val

        kwargs['choices'] = [(True, true_label), (False, false_label)]
        kwargs['coerce'] = bool_coerce
        kwargs['validators'] = [InputRequired()] + kwargs.get('validators', [])

        super(RequiredBoolRadioField, self).__init__(*args, **kwargs)
        self.type = 'RadioField'


class RelationshipFieldBase:
    """Common base for single/multiple select fields that reference ORM relationships.

    Handles one-to-many and many-to-many patterns.

    Note, must use the wtforms-components fields as a base, because we depend on
    lazy-loaded choices. At import time, a field's choices may not be fully available
    in the data. In addition, when pairing the form with an existing record, we need
    to ensure that the option from the record is present even if it would normally
    be filtered (e.g. an inactive option).
    """
    def __init__(self, label=None, orm_cls=None, label_attr=None, fk_attr='id',
                 query_filter=None, coerce=_not_given, **kwargs):
        label = self.field_label_modifier(label)
        self.orm_cls = orm_cls

        self.label_attr = label_attr
        if self.label_attr is None:
            self.label_attr = self.get_best_label_attr()
        if self.label_attr:
            # compute this once and store on the object, rather than for each option
            self.label_attr_name = self.compute_option_attr_name(self.label_attr)

        self.fk_attr = fk_attr
        if self.fk_attr:
            self.fk_attr_name = self.compute_option_attr_name(self.fk_attr)

        self.query_filter = query_filter
        if not self.fk_attr and not coerce:
            def coerce_to_orm_obj(value):
                """Coerce ID to relationship object."""
                # If coming form formdata, we'll get a string ID.
                if isinstance(value, str):
                    return self.orm_cls.get(value)

                # If coming from object data, we'll get an ORM instance.
                return value

            coerce = coerce_to_orm_obj

        super().__init__(label=label, choices=self.get_choices, coerce=coerce, **kwargs)

    def compute_option_attr_name(self, base_attr):
        """To access the label value on an option in the result set, we need to know what
        attribute to use. Based on ``label_attr``, which can be string, SA ORM attr,
        SA hybrid attr, etc., determine a sane attribute."""
        retval = base_attr

        if isinstance(base_attr, sa.orm.InstrumentedAttribute):
            retval = base_attr.name
        else:
            retval = getattr(base_attr, '__name__', str(base_attr))

        return retval

    def field_label_modifier(self, label):
        """Modifies the label to something more human-friendly.

        One-to-many relationships often have a field name like "foo_id", which title cases
        as "Foo Id". Form should only show "Foo", though, so we trim it that way here.
        """
        if label.lower().endswith(' id'):
            return label.rsplit(' ', 1)[0]
        return label

    def build_query(self):
        query = self.query_base()
        query = self.filter_query(query)
        return query

    def query_base(self):
        orm_label_attr = self.label_attr
        if isinstance(self.label_attr, str):
            orm_label_attr = getattr(self.orm_cls, self.label_attr)
        return self.orm_cls.query.order_by(orm_label_attr)

    def get_data_filter(self):
        if self.fk_attr:
            return getattr(self.orm_cls, self.fk_attr_name) == self.data
        else:
            return self.orm_cls.id == self.data.id

    def filter_query(self, query):
        filter_terms = []
        # Get supplied filters.
        if callable(self.query_filter):
            filter_terms.append(self.query_filter(self))
        elif self.query_filter is not None:
            filter_terms.append(self.query_filter)

        # Having an existing value should filter the query.
        if self.data is not None:
            data_filter = self.get_data_filter()
            if data_filter is not None and filter_terms:
                filter_terms.append(data_filter)

        # Apply filter terms with or_, or directly, depending on length.
        if len(filter_terms) == 1:
            query = query.filter(*filter_terms)
        elif len(filter_terms) > 1:
            query = query.filter(sa.sql.or_(*filter_terms))

        return query

    def get_best_label_attr(self):
        if has_column(self.orm_cls, 'label'):
            return 'label'

        if has_column(self.orm_cls, 'name'):
            return 'name'

        return None

    def get_option_label(self, obj):
        if self.label_attr:
            return getattr(obj, self.label_attr_name)

        return str(obj)

    def get_choices(self):
        query = self.build_query()

        def get_value(obj):
            if self.fk_attr:
                return str(getattr(obj, self.fk_attr_name))

            return str(obj.id)

        return [(get_value(obj), self.get_option_label(obj)) for obj in query]

    @property
    def choice_values(self):
        # coerce values used for validation, because the data we're matching will
        # be int type
        return [self.coerce(v) for v in super().choice_values]


class RelationshipField(RelationshipFieldBase, SelectField):
    """SelectField for relationships.

    Args:

        orm_cls (class): Model class of the relationship attribute. Used to query
        records for populating select options.

        label_attr (str): Name of attribute on relationship class to use for select
        option labels.

        fk_attr (str): Optional name of foreign key column of ORM class. If set to
        None, coerce values to instances of ORM class. Otherwise, coerce values to
        the attribute of ORM class the foreign key belongs to. Default is 'id'.

        query_filter (callable): Optional SA query filter criterion for querying select
        options. Can be a function that returns a filter criterion. Function is
        called with the RelationshipField instance it belongs to.

        coerce (callable): Optional function used to coerce form values. By default,
        if fk_attr is set to None, values are coerced to instances of ORM class.
        Otherwise, the default select coersion is applied. Setting this overrides
        default behavior.

        kwargs: Passed to ``SelectField.__init__``.

    Example::
        class Bar(Model):
            name = Column(Unicode(255))
            foos = relationship('Foo', foreign_keys='foos.id')
        class Foo(Model):
            name = Column(Unicode(255))
            bar_id = Column(sa.ForeignKey(Bar.id))
            bar = relationship(Bar, foreign_keys=bar_id)
        class FooForm(ModelForm):
            bar_id = RelationshipField('Bar Label', Bar, 'name')
    """


class RelationshipMultipleField(RelationshipFieldBase, SelectMultipleField):
    """SelectMultipleField for relationships.
    Args:

        orm_cls (class): Model class of the relationship attribute. Used to query
        records for populating select options.

        label_attr (str): Name of attribute on relationship class to use for select
        option labels.

        query_filter (callable): Optional SA query filter criterion for querying select
        options. Can be a function that returns a filter criterion. Function is
        called with the RelationshipField instance it belongs to.

        coerce (callable): Optional function used to coerce form values. By default,
        values are coerced to instances of ORM class. Setting this overrides
        default behavior.

        kwargs: Passed to ``SelectMultipleField.__init__``.

    Example::
        class Bar(Model):
            name = Column(Unicode(255))
            foos = relationship('Foo', foreign_keys='foos.id')
        class Foo(Model):
            name = Column(Unicode(255))
            bar_id = Column(sa.ForeignKey(Bar.id))
            bar = relationship(Bar, foreign_keys=bar_id)
        class BarForm(ModelForm):
            foos = RelationshipMultipleField('Foos', Foo, 'name')
    """
    def __init__(self, label, orm_cls, label_attr=None,
                 query_filter=None, coerce=_not_given, **kwargs):
        coerce = coerce or self.coerce_to_int
        super().__init__(label, orm_cls, label_attr, None, query_filter, coerce, **kwargs)

    def get_data_filter(self):
        if not self.data:
            # Empty set should not add any options. Returning in_ in this case has
            # undesirable results.
            return
        existing_ids = [obj.id for obj in self.data]
        return self.orm_cls.id.in_(existing_ids)

    @staticmethod
    def coerce_to_int(value):
        return int(value)

    def coerce_to_obj(self, value):
        if type(value) in [int, str]:
            return self.orm_cls.get(int(value))
        else:
            return value

    def iter_choices(self):
        selected_ids = [v.id for v in self.data or []]
        for value, label in self.choices():
            selected = selected_ids is not None and self.coerce(value) in selected_ids
            yield (value, label, selected)

    def process_formdata(self, valuelist):
        try:
            self.data = list(self.coerce_to_obj(v) for v in valuelist)
        except ValueError:
            raise ValueError(self.gettext('Invalid Choice: could not coerce'))

    def process_data(self, value):
        try:
            self.data = list(v for v in value)
        except (ValueError, TypeError):
            self.data = None

    def pre_validate(self, form):
        if self.data:
            values = list(c[0] for c in self.choices())
            for d in self.data:
                if str(d.id) not in values:
                    raise ValueError(
                        self.gettext("'%(value)s' is not a valid choice for this field")
                        % dict(value=d)
                    )


class _TypeHintingTextInputBase(wtforms.widgets.TextInput):
    def __init__(self, prefix=None, suffix=None):
        self.prefix = prefix
        self.suffix = suffix
        super().__init__()


class TypeHintingTextInputB3(_TypeHintingTextInputBase):
    """
    A text input widget with a prefix and/or suffix to hint at the expected type or units.
    For use with bootstrap 3
    """
    def __call__(self, field, **kwargs):
        def make_addon(txt):
            return Markup(
                '<span class="input-group-addon">{}</span>'.format(wtforms.widgets.core.escape(txt))
            )

        return Markup(
            '<div class="input-group">{pre}{field}{post}</div>'.format(
                pre=make_addon(self.prefix) if self.prefix else '',
                field=super().__call__(field, **kwargs).__html__(),
                post=make_addon(self.suffix) if self.suffix else ''
            )
        )


class TypeHintingTextInputB4(_TypeHintingTextInputBase):
    """
    A text input widget with a prefix and/or suffix to hint at the expected type or units.
    For use with bootstrap 4
    """
    def __call__(self, field, **kwargs):
        def make_addon(txt, addon_type):
            return Markup(
                '<div class="input-group-{type}">'
                '   <span class="input-group-text">{txt}</span>'
                "</div>".format(type=addon_type, txt=wtforms.widgets.core.escape(txt))
            )

        return Markup(
            '<div class="input-group">{pre}{field}{post}</div>'.format(
                pre=make_addon(self.prefix, "prepend") if self.prefix else "",
                field=super().__call__(field, **kwargs).__html__(),
                post=make_addon(self.suffix, "append") if self.suffix else "",
            )
        )


def _max_for_numeric(digits, scale):
    return Decimal('{}.{}'.format('9' * (digits - scale), '9' * scale))


class FormGenerator(FormGeneratorBase):
    """Model form generator that applies field meta info, provides validators, etc.

    Meta nested class directives (in addition to wtforms-alchemy):
    - include_datetimes_with_default
    - include_required_foreign_keys

    Field class overrides:
    - Use our SelectField instead of the WTForms default
    - Use our RequiredBoolRadioField for non-nullable boolean fields
    - Use RelationshipField for foreign key fields

    Meta info modifiers:
    - Use FieldsMeta.<field_name> if provided
    - Falls back to FieldsMeta.__default__
    - If none of the above, uses a blank FieldsMeta object, which will title case the label.

    Validators:
    - Applies range/scale numeric validators when applicable.
    """
    def __init__(self, form_class):
        super(FormGenerator, self).__init__(form_class)
        self.fields_meta = getattr(self.form_class, 'FieldsMeta', None)

    def skip_column(self, column):
        # Verify the key is not also in exclude=[] so we don't break compatibility with forms
        # that already manually excluded these fields
        if (not self.meta.include_datetimes_with_default
                and isinstance(column.type, ArrowType)
                and column.default
                and column.key not in self.meta.exclude):
            return True

        # include_foreign_keys will pull in all foreign keys on the object. If we want the
        # form to include only required keys, we use include_required_foreign_keys.
        include_required_fks = getattr(self.meta, 'include_required_foreign_keys', False)
        if (include_required_fks and column.foreign_keys and column.nullable is False):
            return False

        return super().skip_column(column)

    def get_field_class(self, column):
        field_cls = super(FormGenerator, self).get_field_class(column)
        if field_cls is SelectFieldBase:
            return SelectField

        is_required_boolean = (field_cls is wtforms.fields.BooleanField
                               and not column.nullable
                               and not column.default)

        if is_required_boolean:
            return RequiredBoolRadioField

        return field_cls

    def get_field_modifier(self, prop):

        # is there an entry in FieldsMeta?
        if hasattr(self.fields_meta, prop.key):
            field_modifier = getattr(self.fields_meta, prop.key)
        else:
            field_modifier = getattr(self.fields_meta, '__default__', _not_given)
            if field_modifier is _not_given:
                field_modifier = FieldMeta

        return field_modifier() if inspect.isclass(field_modifier) else field_modifier

    def create_field(self, prop, column):
        if column.foreign_keys:
            foreign_key = next(iter(column.foreign_keys))
            orm_cls = get_class_by_table(db.Model, foreign_key.column.table)
            validators = self.create_validators(prop, column)
            field = RelationshipField(
                label=to_title_case(str(column.key)),
                orm_cls=orm_cls,
                validators=validators,
            )
        else:
            field = super(FormGenerator, self).create_field(prop, column)

        modifier = self.get_field_modifier(prop)
        if modifier is not None:
            modifier.apply_to_field(field)
        return field

    def create_validators(self, prop, column):
        validators = super(FormGenerator, self).create_validators(prop, column)
        if isinstance(column.type, sa.Numeric) and not isinstance(column.type, sa.Float):
            if column.type.precision is None or column.type.scale is None:
                raise ValueError('Numeric fields must specify precision and scale')
            max_ = _max_for_numeric(column.type.precision, column.type.scale)
            validators.append(NumberRange(min=-max_, max=max_))
            validators.append(NumberScale(column.type.scale))
        return validators

    def length_validator(self, column):
        if isinstance(column.type, sa.types.Enum):
            return None
        return super(FormGenerator, self).length_validator(column)

    def select_field_kwargs(self, column):
        enum_cls = getattr(column.type, 'enum_class', None)
        if enum_cls and issubclass(enum_cls, DBEnum):
            return {
                'coerce': enum_cls.coerce,
                'choices': enum_cls.form_options()
            }
        return super().select_field_kwargs(column)


def field_to_dict(field):
    if isinstance(field, wtforms.fields.FormField):
        return form_fields_to_dict(field)
    if isinstance(field, wtforms.fields.FieldList):
        return [field_to_dict(subfield) for subfield in field]
    return {
        'data': field.data,
        'errors': field.errors,
        'label': field.label.text,
        'required': field.flags.required,
    }


def form_fields_to_dict(form):
    return dict((str(name), field_to_dict(field))
                for name, field in form._fields.items())


___validator_creation_counter = 0


def form_validator(func=None, only_when_fields_valid=False):
    """Decorator used to mark a method as a form level validator.

    :param only_when_fields_valid: Use to disable validator if form already has errors.
    """
    if func is None:
        return functools.partial(form_validator, only_when_fields_valid=only_when_fields_valid)

    @functools.wraps(func)
    def wrapper(form):
        if not only_when_fields_valid or not form.errors:
            return func(form)

    global ___validator_creation_counter
    wrapper.___form_validator = True
    ___validator_creation_counter += 1
    wrapper.___creation_counter = ___validator_creation_counter
    return wrapper


class Form(BaseForm):
    """Base form with a bunch of QoL improvements


    :param _field_order: Relying on the default field ordering can lead to unintuitive forms. It is
        possible to override this by adding the ``_field_order`` class attribute. Set this class
        variable to a tuple or list of field names (addressable via Form._fields['name_of_field'])
        and the form will render in that order. You must include all the fields, except CSRF.
        Forgetting a field or adding one which doesn't exist will cause the form to raise a
        ``ValueError`` and the form will not be rendered.

            class MyForm(Form):
                _field_order = ('field1', 'field2',)

                field1 = String('field1_label')  # Note that we don't use the label in the ordering
                field2 = String()
    """
    _form_ident_enabled = True
    _form_ident_strict = True

    def __init__(self, *args, **kwargs):
        super(Form, self).__init__(*args, **kwargs)
        self._form_level_errors = []
        self._errors = None
        self.after_init(args, kwargs)

    def __init_subclass__(cls):
        cls.add_form_ident()
        super().__init_subclass__()

    def __iter__(self):
        custom_field_order = getattr(self, '_field_order', None)

        if custom_field_order is None:
            return super().__iter__()

        order = []
        if hasattr(self, 'csrf_token'):
            order.append('csrf_token')
        if self._form_ident_enabled:
            order.append(self._form_ident_key())
        order.extend(list(custom_field_order))

        declared = set(self._fields.keys())
        ordered = set(order)

        if declared != ordered:
            not_ordered = declared - ordered
            extra_ordered = ordered - declared
            raise ValueError(
                'Custom field ordering for {} is incorrect.'.format(self.__class__.__name__),
                ' Missing fields: {} '.format(not_ordered),
                ' Extra fields: {} '.format(extra_ordered),
            )

        return (self._fields[f] for f in order)

    def after_init(self, args, kwargs):
        """Hook for providing customization on the form after fields are initialized."""
        pass

    def fields_todict(self):
        """Turns a form into dicts and lists with both data and errors for each field."""
        return form_fields_to_dict(self)

    def validate(self):
        """Applies validators and returns bool.

        Methods decorated as form-level validators are run after WTForms generic validation.
        """
        fields_valid = super(Form, self).validate()

        form_validators = {}
        # Traverse the MRO so we can get validators in parent classes.
        # Do so in reverse order so child classes can override parents' validators.
        # WTForms will not include the methods on form instances so we get them from the classes.
        for cls in reversed(self.__class__.__mro__):
            cls_validators = {
                name: attr for name, attr in cls.__dict__.items()
                if getattr(attr, '___form_validator', False)
            }
            form_validators.update(cls_validators)

        for validator in sorted(form_validators.values(), key=attrgetter('___creation_counter')):
            try:
                validator(self)
            except StopValidation as e:
                if e.args and e.args[0]:
                    self.form_errors.append(e.args[0])
                break
            except ValueError as e:
                self.form_errors.append(e.args[0])

        return fields_valid and not self.form_errors

    @property
    def field_errors(self):
        """Field-level validator errors come from WTForms' errors."""
        warnings.warn(
            'WTForms has form-level validation now, use form.errors instead', DeprecationWarning, 2
        )
        return self.errors

    @classmethod
    def add_form_ident(cls):
        # may need to clean up from a superclass init, so we have fresh config here
        key = cls._form_ident_key()
        if hasattr(cls, key):
            setattr(cls, key, None)

        if not cls._form_ident_enabled:
            return

        if key.startswith('_'):
            raise Exception('Cannot start form ident name with "_", since WTForms will ignore')

        validators = []
        value = cls._form_ident_value()
        if cls._form_ident_strict:
            validators.append(AnyOf([value]))

        setattr(
            cls,
            key,
            wtforms.fields.HiddenField(
                default=value,
                validators=validators,
            )
        )

    @classmethod
    def _form_ident_key(cls):
        """Field name to embed as a hidden value for form identification. Default is keg_form_ident.

        Note: this cannot start with an underscore, or WTForms will ignore the field.
        """
        return 'keg_form_ident'

    @classmethod
    def _form_ident_value(cls):
        """Field value to embed for form identification. Default is class name converted to
        dash notation."""
        return case_cw2dash(cls.__name__)


BaseModelFormMeta = model_form_meta_factory()


class ModelFormMeta(BaseModelFormMeta):
    """Base model form metaclass that handles nested inheritance issues.

    The default metaclass here will handle the nested Meta class. A form
    subclass with a Meta nested class will treat the form's superclass' Meta
    as a parent.

    This metaclass does the same thing for FieldsMeta, allowing superclasses
    to define a FieldsMeta that may reasonably be passed down to the subclass.
    """
    def __init__(cls, *args, **kwargs):
        bases = []
        for class_ in cls.__mro__:
            if 'FieldsMeta' in class_.__dict__:
                bases.append(getattr(class_, 'FieldsMeta'))

        if object not in bases:
            bases.append(object)

        cls.FieldsMeta = type('FieldsMeta', tuple(bases), {})

        BaseModelFormMeta.__init__(cls, *args, **kwargs)


BaseModelForm = model_form_factory(Form, meta=ModelFormMeta, form_generator=FormGenerator)


class ModelForm(BaseModelForm):
    """Base model-generated form class that applies KegElements generator and meta."""
    @classmethod
    def get_session(cls):
        return db.session
